@@ -9,19 +9,20 @@
 #include "event.h"
 #include "convert.h"
 #include "dictstat.h"
-#include "thread.h"
 #include "rp.h"
 #include "rp_cpu.h"
 #include "shared.h"
 #include "wordlist.h"
+#include "emu_inc_hash_sha1.h"
 
 size_t convert_from_hex (hashcat_ctx_t *hashcat_ctx, char *line_buf, const size_t line_len)
 {
+  const hashconfig_t   *hashconfig   = hashcat_ctx->hashconfig;
   const user_options_t *user_options = hashcat_ctx->user_options;
 
   if (line_len & 1) return (line_len); // not in hex
 
-  if (user_options->hex_wordlist == true)
+  if (hashconfig->opts_type & OPTS_TYPE_PT_HEX)
   {
     size_t i, j;
 
@@ -48,7 +49,7 @@ size_t convert_from_hex (hashcat_ctx_t *hashcat_ctx, char *line_buf, const size_
   return (line_len);
 }
 
-int load_segment (hashcat_ctx_t *hashcat_ctx, FILE *fd)
+int load_segment (hashcat_ctx_t *hashcat_ctx, HCFILE *fp)
 {
   wl_data_t *wl_data = hashcat_ctx->wl_data;
 
@@ -56,7 +57,7 @@ int load_segment (hashcat_ctx_t *hashcat_ctx, FILE *fd)
 
   wl_data->pos = 0;
 
-  wl_data->cnt = hc_fread (wl_data->buf, 1, wl_data->incr - 1000, fd);
+  wl_data->cnt = hc_fread (wl_data->buf, 1, wl_data->incr - 1000, fp);
 
   wl_data->buf[wl_data->cnt] = 0;
 
@@ -64,7 +65,7 @@ int load_segment (hashcat_ctx_t *hashcat_ctx, FILE *fd)
 
   if (wl_data->buf[wl_data->cnt - 1] == '\n') return 0;
 
-  while (!feof (fd))
+  while (!hc_feof (fp))
   {
     if (wl_data->cnt == wl_data->avail)
     {
@@ -73,7 +74,7 @@ int load_segment (hashcat_ctx_t *hashcat_ctx, FILE *fd)
       wl_data->avail += wl_data->incr;
     }
 
-    const int c = fgetc (fd);
+    const int c = hc_fgetc (fp);
 
     if (c == EOF) break;
 
@@ -171,7 +172,7 @@ void get_next_word_std (char *buf, u64 sz, u64 *len, u64 *off)
   *len = sz;
 }
 
-void get_next_word (hashcat_ctx_t *hashcat_ctx, FILE *fd, char **out_buf, u32 *out_len)
+void get_next_word (hashcat_ctx_t *hashcat_ctx, HCFILE *fp, char **out_buf, u32 *out_len)
 {
   user_options_t       *user_options       = hashcat_ctx->user_options;
   user_options_extra_t *user_options_extra = hashcat_ctx->user_options_extra;
@@ -188,7 +189,13 @@ void get_next_word (hashcat_ctx_t *hashcat_ctx, FILE *fd, char **out_buf, u32 *o
 
     wl_data->pos += off;
 
+    // do the on-the-fly hex decode using original buffer
+    // this is safe as length only decreases in size
+
+    len = (u32) convert_from_hex (hashcat_ctx, ptr, len);
+
     // do the on-the-fly encoding
+    // needs to write into new buffer because size case both decrease and increase
 
     if (wl_data->iconv_enabled == true)
     {
@@ -205,6 +212,8 @@ void get_next_word (hashcat_ctx_t *hashcat_ctx, FILE *fd, char **out_buf, u32 *o
       len = HCBUFSIZ_TINY - iconv_sz;
     }
 
+    // this is only a test for length, not writing into output buffer
+
     if (run_rule_engine (user_options_extra->rule_len_l, user_options->rule_buf_l))
     {
       if (len >= RP_PASSWORD_SIZE) continue;
@@ -218,7 +227,7 @@ void get_next_word (hashcat_ctx_t *hashcat_ctx, FILE *fd, char **out_buf, u32 *o
       if (rule_len_out < 0) continue;
     }
 
-    if (len >= PW_MAX) continue;
+    if (len > PW_MAX) continue;
 
     *out_buf = ptr;
     *out_len = (u32) len;
@@ -226,16 +235,61 @@ void get_next_word (hashcat_ctx_t *hashcat_ctx, FILE *fd, char **out_buf, u32 *o
     return;
   }
 
-  if (feof (fd))
+  if (hc_feof (fp))
   {
     fprintf (stderr, "BUG feof()!!\n");
 
     return;
   }
 
-  load_segment (hashcat_ctx, fd);
+  load_segment (hashcat_ctx, fp);
 
-  get_next_word (hashcat_ctx, fd, out_buf, out_len);
+  get_next_word (hashcat_ctx, fp, out_buf, out_len);
+}
+
+void pw_pre_add (hc_device_param_t *device_param, const u8 *pw_buf, const int pw_len, const u8 *base_buf, const int base_len, const int rule_idx)
+{
+  if (device_param->pws_pre_cnt < device_param->kernel_power)
+  {
+    pw_pre_t *pw_pre = device_param->pws_pre_buf + device_param->pws_pre_cnt;
+
+    memcpy (pw_pre->pw_buf, pw_buf, pw_len);
+
+    pw_pre->pw_len = pw_len;
+
+    if (base_buf != NULL)
+    {
+      memcpy (pw_pre->base_buf, base_buf, base_len);
+
+      pw_pre->base_len = base_len;
+    }
+
+    pw_pre->rule_idx = rule_idx;
+
+    device_param->pws_pre_cnt++;
+  }
+  else
+  {
+    fprintf (stdout, "BUG pw_pre_add()!!\n");
+
+    return;
+  }
+}
+
+void pw_base_add (hc_device_param_t *device_param, pw_pre_t *pw_pre)
+{
+  if (device_param->pws_base_cnt < device_param->kernel_power)
+  {
+    memcpy (device_param->pws_base_buf + device_param->pws_base_cnt, pw_pre, sizeof (pw_pre_t));
+
+    device_param->pws_base_cnt++;
+  }
+  else
+  {
+    fprintf (stderr, "BUG pw_base_add()!!\n");
+
+    return;
+  }
 }
 
 void pw_add (hc_device_param_t *device_param, const u8 *pw_buf, const int pw_len)
@@ -273,7 +327,7 @@ void pw_add (hc_device_param_t *device_param, const u8 *pw_buf, const int pw_len
   }
 }
 
-int count_words (hashcat_ctx_t *hashcat_ctx, FILE *fd, const char *dictfile, u64 *result)
+int count_words (hashcat_ctx_t *hashcat_ctx, HCFILE *fp, const char *dictfile, u64 *result)
 {
   combinator_ctx_t     *combinator_ctx     = hashcat_ctx->combinator_ctx;
   hashconfig_t         *hashconfig         = hashcat_ctx->hashconfig;
@@ -287,9 +341,9 @@ int count_words (hashcat_ctx_t *hashcat_ctx, FILE *fd, const char *dictfile, u64
 
   dictstat_t d;
 
-  d.cnt = 0;
+  memset (&d, 0, sizeof (d));
 
-  if (fstat (fileno (fd), &d.stat))
+  if (fstat (hc_fileno (fp), &d.stat))
   {
     *result = 0;
 
@@ -315,8 +369,8 @@ int count_words (hashcat_ctx_t *hashcat_ctx, FILE *fd, const char *dictfile, u64
   memset (d.encoding_from, 0, sizeof (d.encoding_from));
   memset (d.encoding_to,   0, sizeof (d.encoding_to));
 
-  strncpy (d.encoding_from, user_options->encoding_from, sizeof (d.encoding_from));
-  strncpy (d.encoding_to,   user_options->encoding_to,   sizeof (d.encoding_to));
+  strncpy (d.encoding_from, user_options->encoding_from, sizeof (d.encoding_from) - 1);
+  strncpy (d.encoding_to,   user_options->encoding_to,   sizeof (d.encoding_to)   - 1);
 
   if (d.stat.st_size == 0)
   {
@@ -325,6 +379,21 @@ int count_words (hashcat_ctx_t *hashcat_ctx, FILE *fd, const char *dictfile, u64
     return 0;
   }
 
+  const size_t dictfile_len = strlen (dictfile);
+
+  u32 *dictfile_padded = (u32 *) hcmalloc (dictfile_len + 64); // padding required for sha1_update()
+
+  memcpy (dictfile_padded, dictfile, dictfile_len);
+
+  sha1_ctx_t sha1_ctx;
+  sha1_init   (&sha1_ctx);
+  sha1_update (&sha1_ctx, dictfile_padded, dictfile_len);
+  sha1_final  (&sha1_ctx);
+
+  hcfree (dictfile_padded);
+
+  memcpy (d.hash_filename, sha1_ctx.h, 16);
+
   const u64 cached_cnt = dictstat_find (hashcat_ctx, &d);
 
   if (run_rule_engine (user_options_extra->rule_len_l, user_options->rule_buf_l) == 0)
@@ -381,9 +450,9 @@ int count_words (hashcat_ctx_t *hashcat_ctx, FILE *fd, const char *dictfile, u64
   u64 cnt  = 0;
   u64 cnt2 = 0;
 
-  while (!feof (fd))
+  while (!hc_feof (fp))
   {
-    load_segment (hashcat_ctx, fd);
+    load_segment (hashcat_ctx, fp);
 
     comp += wl_data->cnt;
 
@@ -400,6 +469,11 @@ int count_words (hashcat_ctx_t *hashcat_ctx, FILE *fd, const char *dictfile, u64
 
       i += off;
 
+      // do the on-the-fly hex decode using original buffer
+      // this is safe as length only decreases in size
+
+      len = (u32) convert_from_hex (hashcat_ctx, ptr, len);
+
       // do the on-the-fly encoding
 
       if (wl_data->iconv_enabled == true)
@@ -432,7 +506,7 @@ int count_words (hashcat_ctx_t *hashcat_ctx, FILE *fd, const char *dictfile, u64
 
       cnt2++;
 
-      if (len >= PW_MAX) continue;
+      if (len > PW_MAX) continue;
 
       d.cnt++;
 
@@ -516,7 +590,7 @@ int wl_data_init (hashcat_ctx_t *hashcat_ctx)
   if (user_options->benchmark      == true) return 0;
   if (user_options->example_hashes == true) return 0;
   if (user_options->left           == true) return 0;
-  if (user_options->opencl_info    == true) return 0;
+  if (user_options->backend_info   == true) return 0;
   if (user_options->usage          == true) return 0;
   if (user_options->version        == true) return 0;
 
@@ -539,7 +613,7 @@ int wl_data_init (hashcat_ctx_t *hashcat_ctx)
     wl_data->func = get_next_word_uc;
   }
 
-  if (hashconfig->hash_mode == 3000) // yes that's fine that way
+  if (hashconfig->opts_type & OPTS_TYPE_PT_LM)
   {
     wl_data->func = get_next_word_lm;
   }
